cdef extern from "<uchardet/uchardet.h>" nogil:
    ctypedef struct uchardet
    ctypedef uchardet* uchardet_t

    uchardet_t uchardet_new()
    void uchardet_delete(uchardet_t ud)
    int uchardet_handle_data(uchardet_t ud, const char* data, size_t len)
    void uchardet_data_end(uchardet_t ud)
    void uchardet_reset(uchardet_t ud)
    const char * uchardet_get_charset(uchardet_t ud)



cdef extern from "<dlfcn.h>" nogil:
    const int RTLD_LAZY
    const int RTLD_NOW
    void* dlopen(const char* filename, int flags)
    void* dlsym(void* handle, const char* symbol);
    int dlclose(void* handle)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

import typing as t

from libcpp.algorithm cimport pop_heap, push_heap
cimport cython
from cython.operator cimport preincrement as preinc
from cpython.unicode cimport Py_UNICODE_ISALPHA, Py_UNICODE_ISSPACE


@cython.wraparound(False)
cdef lang_vec8_t str_to_vec(str train_text, size_t vec_len=LANG_VEC_SIZE):
    cdef long hash2
    cdef long hash3
    cdef long hash4
    cdef long hash5
    cdef Py_UCS4[2] ngram2 = [0, 0]
    cdef Py_UCS4[3] ngram3 = [0, 0, 0]
    cdef Py_UCS4[4] ngram4 = [0, 0, 0, 0]
    cdef Py_UCS4[5] ngram5 = [0, 0, 0, 0, 0]

    cdef lang_vec32_t count_vec32
    count_vec32.resize(vec_len)

    cdef bint prev_is_space = False
    cdef size_t i = 0
    cdef Py_UCS4 uchar
    for uchar in train_text:
        if Py_UNICODE_ISALPHA(uchar):
            prev_is_space = False
        elif Py_UNICODE_ISSPACE(uchar):
            if prev_is_space:
                continue
            prev_is_space = True
            uchar = 0x20
        else:
            prev_is_space = False
            continue

        # Shift n-gram buffers
        if i > 0:
            ngram2[0] = ngram2[1]
            shiftleft(ngram3, 3)
            shiftleft(ngram4, 4)
            shiftleft(ngram5, 5)
        ngram2[1] = uchar
        ngram3[2] = uchar
        ngram4[3] = uchar
        ngram5[4] = uchar

        if vec_len > 255:
            preinc(count_vec32[hash_fnv8_single(uchar)])
            if i >= 1:
                preinc(count_vec32[hash_fnv8(ngram2, 2)])
            if i >= 2:
                preinc(count_vec32[hash_fnv8(ngram3, 3)])
            if i >= 3:
                preinc(count_vec32[hash_fnv8(ngram4, 4)])
            if i >= 4:
                preinc(count_vec32[hash_fnv8(ngram5, 5)])
        else:
            preinc(count_vec32[hash_fnv8_single(uchar) % vec_len])
            if i >= 1:
                preinc(count_vec32[hash_fnv8(ngram2, 2) % vec_len])
            if i >= 2:
                preinc(count_vec32[hash_fnv8(ngram3, 3) % vec_len])
            if i >= 3:
                preinc(count_vec32[hash_fnv8(ngram4, 4) % vec_len])
            if i >= 4:
                preinc(count_vec32[hash_fnv8(ngram5, 5) % vec_len])

        preinc(i)

    # Normalize vector
    cdef size_t j
    cdef lang_vec8_t lang_vec
    lang_vec.resize(vec_len)
    if i > 0:
        for j in range(vec_len):
            lang_vec[j] = <uint8_t>min(255u, count_vec32[j] * 256u / i)

    return lang_vec


cdef inline bint lang_rank_greater(const lang_rank_t& a, const lang_rank_t& b):
    return a.rank > b.rank


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef detect_fast(str text, size_t cutoff=1200, size_t n_results=1, langs=None):
    """
    detect_fast(text, cutoff=1200, n_results=1, langs=None)
    
    Perform a very fast (linear-time) language detection on the input string.
    
    The output is a tuple of the detected language name and the calculated
    out-of-place rank, which indicates how far the given text is from the closest-matching
    language profile. The higher the rank, the less accurate the detection is. Values
    above 1200 are usually false results.
    
    The given Unicode string should be in composed normal form (NFC) for the best results.
    
    :param text: input text
    :type text: str
    :param cutoff: OOP rank cutoff after which to return ``"unknown"``
    :type cutoff: int
    :param n_results: if this is greater than one, a list of the ``n_results`` best matches will be returned
    :type n_results: int
    :param langs: restrict detection to these languages
    :type langs: list[str]
    :return: tuple of the detected language (or ``"unknown"``) and its out-of-place rank
    :rtype: (str, int) | list[(str, int)]
    """
    cdef lang_vec8_t text_vec = str_to_vec(text, LANG_VEC_SIZE)
    cdef size_t text_len = len(text)
    cdef size_t min_rank = <size_t>-1
    cdef const char* lang = NULL
    cdef vector[lang_rank_t] predicted
    cdef uint32_t i
    cdef uint32_t rank

    if langs:
        langs = set(langs)

    for i in range(N_LANGS):
        if langs is not None and LANGS[i].lang.decode() not in langs:
            continue

        rank = cmp_oop_ranks(text_vec.data(), LANGS[i].vec, LANG_VEC_SIZE)
        # Bias rank by position in the language list on short texts with high uncertainty
        if rank > 500 and text_len < 150:
            rank += min(50u, i * 3u)
        if rank > cutoff:
            continue

        if n_results == 1 and rank < min_rank:
            min_rank = rank
            lang = LANGS[i].lang
        elif n_results > 1:
            predicted.push_back([rank, LANGS[i].lang])
            push_heap(predicted.begin(), predicted.end(), &lang_rank_greater)

    if n_results == 1:
        if lang == NULL:
            return "unknown", 0
        return lang.decode(), min_rank

    result_list = []
    for i in range(min(n_results, predicted.size())):
        pop_heap(predicted.begin(), predicted.end(), &lang_rank_greater)
        result_list.append((predicted.back().lang.decode(), predicted.back().rank))
        predicted.pop_back()

    return result_list


def supported_langs():
    """
    supported_langs()

    Get a list of all languages that are supported by the fast language detector.

    :return: list of supported languages
    :rtype: list[str]
    """

    cdef size_t i
    langs = []
    for i in range(N_LANGS):
        langs.append(LANGS[i].lang.decode())
    return sorted(langs)


@cython.wraparound(False)
cpdef train_language_examples(examples, size_t vec_len=LANG_VEC_SIZE):
    """
    train_language_examples(examples, vec_len=256)

    Train a language vector for fast language detection on a list of example texts.

    :param examples: list of example texts for this language
    :type examples: t.Iterable[str]
    :param vec_len: output vector length
    :type vec_len: int
    :return: vector of trained values
    :rtype: list[int]
    """
    cdef lang_vec32_t agg_vec
    agg_vec.resize(vec_len)

    cdef lang_vec8_t tmp_vec
    cdef uint32_t example_count = 0
    cdef uint32_t i
    for text in examples:
        tmp_vec = str_to_vec(text, vec_len)
        for i in range(tmp_vec.size()):
            agg_vec[i] += tmp_vec[i]
        example_count += 1

    cdef lang_vec8_t agg_vec8
    agg_vec8.resize(agg_vec.size())
    for i in range(agg_vec.size()):
        agg_vec8[i] = <uint8_t>min(255u, agg_vec[i] / example_count)

    return agg_vec8



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
from collections import defaultdict
import json
import hashlib
import os
from math import fsum
import re
import time
import tempfile
import unicodedata
import urllib.request
from urllib.error import HTTPError
import sys

from fastwarc import ArchiveIterator, WarcRecordType
from resiliparse.parse.encoding import detect_encoding, bytes_to_str
from resiliparse.parse.html import HTMLTree
import resiliparse.parse.lang as rlang


# Optional 'cli' dependencies
MISSING_DEPS = False
try:
    import click
except ModuleNotFoundError:
    print('Missing dependency: click', file=sys.stderr)
    print('Install Resiliparse with the "cli" flag: pip install "resiliparse[cli]"', file=sys.stderr)
    sys.exit(1)

try:
    from tqdm import tqdm
    import joblib
except ModuleNotFoundError:
    MISSING_DEPS = True


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
def main():
    """Resiliparse Command Line Interface."""

    if MISSING_DEPS:
        click.echo('Missing CLI dependencies!', err=True)
        click.echo('Install Resiliparse with the "cli" flag: pip install "resiliparse[cli]"', err=True)
        sys.exit(1)


@main.group()
def encoding():
    """Encoding module tools."""
    pass


@encoding.command()
def download_whatwg_mapping():
    """
    Download WHATWG encoding mapping.

    Download the current WHATWG encoding mapping, parse and transform it, and
    then print it as a copyable Python dict.
    """

    encodings = json.load(urllib.request.urlopen('https://encoding.spec.whatwg.org/encodings.json'))

    enc_mapped = {}
    for enc in encodings:
        for enc2 in enc['encodings']:
            n = enc2['name'].lower()
            try:
                if n == 'iso-8859-8-i':
                    n = 'iso-8859-8'
                    enc_mapped[n.lower()] = n
                elif n == 'windows-874':
                    n = 'iso-8859-11'
                n = codecs.lookup(n).name
                enc_mapped[n] = n
                for l in enc2['labels']:
                    enc_mapped[l] = n
            except LookupError:
                click.echo(f'skipped {enc2["name"]}', err=True)
    enc_mapped = dict((k, enc_mapped[k]) for k in sorted(enc_mapped))

    click.echo(enc_mapped)


@main.group()
def html():
    """HTML module tools."""
    pass


@html.command()
@click.argument('warc_file', type=click.Path(dir_okay=False, exists=True))
def benchmark(warc_file):
    """
    Benchmark Resiliparse HTML parser.

    Benchmark Resiliparse HTML parsing by extracting the titles from all HTML pages in a WARC file.

    You can compare the performance to Selectolax (both the old MyHTML and the new Lexbor engine) and
    BeautifulSoup4 by installing the PyPi packages ``selectolax`` and ``beautifulsoup4``. Install
    Resiliparse with the ``cli-benchmark`` flag to install all optional third-party dependencies automatically.

    See :ref:`Resiliparse HTML Parser Benchmarks <parse-html-benchmark>` for more details
    and example benchmarking results.
    """

    print('HTML parser benchmark <title> extraction:')
    print('=========================================')

    start = time.monotonic()
    i = 0
    for i, record in enumerate(tqdm(ArchiveIterator(open(warc_file, 'rb'), record_types=WarcRecordType.response),
                                    desc=f'Benchmarking Resiliparse (Lexbor)', unit=' docs',
                                    leave=False, mininterval=0.3)):
        content = record.reader.read()
        # noinspection PyStatementEffect
        HTMLTree.parse_from_bytes(content, detect_encoding(content)).title
    t = time.monotonic() - start
    print(f'Resiliparse (Lexbor):  {i} documents in {t:.2f}s ({i / t:.2f} documents/s)')

    try:
        from selectolax.lexbor import LexborHTMLParser as SelectolaxLexbor
        start = time.monotonic()
        i = 0
        for i, record in enumerate(tqdm(ArchiveIterator(open(warc_file, 'rb'), record_types=WarcRecordType.response),
                                        desc=f'Benchmarking Selectolax (Lexbor)', unit=' docs',
                                        leave=False, mininterval=0.3)):
            content = record.reader.read()
            SelectolaxLexbor(bytes_to_str(content, detect_encoding(content))).css_first('title')
        t = time.monotonic() - start
        print(f'Selectolax (Lexbor):   {i} documents in {time.monotonic() - start:.2f}s ({i / t:.2f} documents/s)')

        from selectolax.parser import HTMLParser as SelectolaxMyHTML
        start = time.monotonic()
        i = 0
        for i, record in enumerate(tqdm(ArchiveIterator(open(warc_file, 'rb'), record_types=WarcRecordType.response),
                                        desc=f'Benchmarking Selectolax (MyHTML)', unit=' docs',
                                        leave=False, mininterval=0.3)):
            content = record.reader.read()
            SelectolaxMyHTML(bytes_to_str(content, detect_encoding(content))).css_first('title')
        t = time.monotonic() - start
        print(f'Selectolax (MyHTML):   {i} documents in {time.monotonic() - start:.2f}s ({i / t:.2f} documents/s)')
    except ModuleNotFoundError:
        print('[Not benchmarking Selectolax, because it is not installed.')

    try:
        from bs4 import BeautifulSoup
        start = time.monotonic()
        for i, record in enumerate(tqdm(ArchiveIterator(open(warc_file, 'rb'), record_types=WarcRecordType.response),
                                        desc=f'Benchmarking BeautifulSoup4 (lxml)', unit=' docs',
                                        leave=False, mininterval=0.3)):
            content = record.reader.read()
            # noinspection PyStatementEffect
            BeautifulSoup(bytes_to_str(content, detect_encoding(content)), 'lxml').title
        t = time.monotonic() - start
        print(f'BeautifulSoup4 (lxml): {i} documents in {time.monotonic() - start:.2f}s ({i / t:.2f} documents/s)')
    except ModuleNotFoundError:
        print('[Not benchmarking BeautifulSoup4, because it is not installed.')


@main.group()
def lang():
    """Language module tools."""
    pass


DEFAULT_WIKI_LANGS = ('en,ru,zh,de,vi,fr,sv,uk,pt,fa,sr,it,pl,ja,nl,es,ca,sh,ko,tr,fi,cs,no,he,id,hu,el,ro,th,ar,bg,hy,'
                      'da,bn,cy,eo,hr,ka,eu,gl,hi,be,sk,ur,ms,ml,mk,et,te,lt,ce,az,tt,kk,sl,ta,lv,uz,mr,tl,my,oc,vo,tg,'
                      'bs,ne,is,ku,mg,pa,nn,sw,ga,fy,la,li,af,io,qu,os,sa,gd,yo,wa,sq,zu,ve,yi,wo,lo,kw,ln,gu,ti,ha,mn,'
                      'ky,su,lg,ba,kn,ee,ie,se,sc,or,am,fo,kv,ps,sm,za,lb,dv,rm,br,ch,cu,iu,sn,an,jv,mt,mi,om,ki,km,ia,'
                      'ss,si,rw,bo,st,so,kl,sg,na,ht,cr,as,av,rn,sd,bh,tw,tk,ig,ug,ff,ny,gn,to,cv,ks,xh,bi,dz,pi,ay,co,'
                      'ty,fj,ik,nv,ak,ab,gv,tn,kg,bm,ts')


@lang.command()
@click.argument('dumpdate')
@click.option('-l', '--langs', help='Comma-separated list of languages to download', default=DEFAULT_WIKI_LANGS)
@click.option('-o', '--outdir', help='Output directory', default='wikidumps', type=click.Path(file_okay=False))
@click.option('-j', '--jobs', help='Parallel download jobs (3 is the Wikimedia rate limit)', default=3, type=int)
def download_wiki_dumps(dumpdate, langs, outdir, jobs):
    """
    Download Wikipedia dumps for language detection.

    Download the first Wikipedia article multistream part for each of the specified languages.

    The downloaded dumps can then be extracted with `Wikiextractor <https://github.com/attardi/wikiextractor>`_.
    """

    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    def dl(l):
        try:
            with urllib.request.urlopen(f'https://dumps.wikimedia.org/{l}wiki/{dumpdate}/dumpstatus.json') as metaf:
                metadata = json.load(metaf)

            if 'articlesmultistreamdump' not in metadata['jobs']:
                return
            metadata = next(iter(metadata['jobs']['articlesmultistreamdump']['files'].values()))
            url = metadata['url']
            md5sum = metadata['md5']
            total_bytes = metadata['size']
            md5hash = hashlib.md5()
            outfilepath = os.path.join(outdir, f'{l}wiki.{url.split(".")[-1]}')
            with open(outfilepath, 'wb') as outf:
                    with urllib.request.urlopen(f'https://dumps.wikimedia.org{url}') as dumpf:
                        blocksize = 4096
                        with tqdm(desc=f'Downloading {l}wiki', unit='b', unit_scale=True, total=total_bytes,
                                  leave=False) as dlprog:
                            while buf := dumpf.read(blocksize):
                                md5hash.update(buf)
                                outf.write(buf)
                                dlprog.update(len(buf))
            if md5hash.hexdigest() != md5sum:
                click.echo(f'Output "{os.path.basename(outfilepath)}" corrupted, deleting it.', err=True)
                os.unlink(outfilepath)
        except HTTPError as e:
            click.echo(f'Error downloading {l}wiki: {e}', err=True)

    joblib.Parallel(n_jobs=jobs, verbose=0, prefer='threads')(joblib.delayed(dl)(l.strip()) for l in langs.split(','))


@lang.command()
@click.argument('indir')
@click.argument('outdir')
@click.option('--val-size', help='Portion of the data to use for validation', default=5, type=int)
@click.option('--test-size', help='Portion of the data to use for testing', default=5, type=int)
@click.option('--min-examples', help='Minimum number of examples per language', default=10000,
              show_default=True, type=int)
@click.option('-j', '--jobs', help='Parallel jobs', default=4, type=int)
def create_dataset(indir, outdir, val_size, test_size, min_examples, jobs):
    """
    Create a language detection dataset.

    Create a language detection dataset from a set of extracted Wikipedia article dumps.

    Expected is a directory containing one subdirectory per language (with the language name,
    e.g. "en" or "enwiki") with any number of subdirectories and ``wiki_*`` plaintext files. Use
    `Wikiextractor <https://github.com/attardi/wikiextractor>`_ for creating the plaintext
    directories for each language.
    
    Empty lines and ``<doc>`` tags will be stripped from the plaintext, otherwise the texts
    are expected to be clean already.

    The created dataset will consist of one directory for each language, each containing three files
    for train, validation, and test with one example per line. The order of the lines is randomized.
    """
    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    langdirs = [os.path.join(indir, d) for d in os.listdir(indir) if os.path.isdir(os.path.join(indir, d))]
    joblib.Parallel(n_jobs=jobs, verbose=10, backend='multiprocessing')(
        joblib.delayed(_process_raw_lang_dir)(l, outdir, val_size, test_size, min_examples) for l in langdirs)


def _process_raw_lang_dir(indir, outdir_base, val_size, test_size, min_examples):
    lang_name = os.path.basename(indir).replace('wiki', '')
    outdir = os.path.join(outdir_base, lang_name)

    train_name = os.path.join(outdir, 'train.txt')
    val_name = os.path.join(outdir, 'val.txt')
    test_name = os.path.join(outdir, 'test.txt')

    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    elif os.path.isfile(train_name) and os.path.getsize(train_name) > 0:
        click.echo(f'Skipping {lang_name}, since dataset for language already present.', err=True)
        return

    def dir_walk():
        for d, f in ((d[0], d[2]) for d in os.walk(indir)):
            for fi in f:
                yield os.path.join(d, fi)

    def hash_fname(base_dir, hash_hex):
        d = os.path.join(base_dir, hash_hex[:2], hash_hex[2:4])
        os.makedirs(d, exist_ok=True)
        return os.path.join(d, hash_hex)

    wiki_markup_re = re.compile(r'(\[\[|\]\])')
    try:
        with tempfile.TemporaryDirectory(dir=outdir) as tempdir:
            line_hashes = set()
            click.echo(f'Parsing input files for {lang_name}...', err=True)
            for infile_name in dir_walk():
                if not os.path.basename(infile_name).startswith('wiki_'):
                    continue

                with open(infile_name, 'r') as infile:
                    for line in infile:
                        line = wiki_markup_re.sub('', unicodedata.normalize('NFKC', line))
                        if len(line) < 200 or line.startswith('<doc id=') or line.startswith('</doc>'):
                            continue

                        h = hashlib.sha1()
                        h.update(line.encode())
                        with open(hash_fname(tempdir, h.hexdigest()), 'w') as outfile:
                            outfile.write(line)
                        line_hashes.add(h.digest())

            if len(line_hashes) < min_examples:
                click.echo(f'Not enough examples for creating a dataset for {lang_name} ({len(line_hashes)}).',
                           err=True)
                return

            test_end = int(len(line_hashes) * test_size * 0.01)
            val_end = test_end + int(len(line_hashes) * val_size * 0.01)

            with open(test_name, 'w') as testf, open(val_name, 'w') as valf, open(train_name, 'w') as trainf:
                click.echo(f'Writing examples for {lang_name}...', err=True)
                for i, h in enumerate(line_hashes):
                    hname = hash_fname(tempdir, h.hex())
                    if i < test_end:
                        testf.write(open(hname).read())
                    elif test_end < i < val_end:
                        valf.write(open(hname).read())
                    else:
                        trainf.write(open(hname).read())
                    os.unlink(hname)
            click.echo(f'Examples for {lang_name} finished.', err=True)
    finally:
        if not os.listdir(outdir):
            os.rmdir(outdir)


# Wikipedia languages ordered by number of users
# Data taken from https://en.wikipedia.org/wiki/List_of_Wikipedias
_WIKI_BIAS = ['en', 'es', 'fr', 'de', 'zh', 'ru', 'pt', 'it', 'ar', 'ja', 'tr', 'id', 'nl', 'simple', 'pl', 'fa', 'he',
              'vi', 'sv', 'ko', 'hi', 'uk', 'ro', 'cs', 'no', 'fi', 'hu', 'da', 'th', 'ca', 'bn', 'el', 'bg', 'sr',
              'ms', 'hr', 'az', 'zh-yue', 'sk', 'sl', 'ta', 'eo', 'sh', 'arz', 'lt', 'et', 'ml', 'la', 'af', 'mr', 'bs',
              'sq', 'ur', 'ka', 'eu', 'gl', 'tl', 'nn', 'hy', 'ang', 'kk', 'be', 'te', 'lv', 'mk', 'my', 'ast',
              'zh-classical', 'sco', 'als', 'ceb', 'is', 'wuu', 'mn', 'be-tarask', 'kn', 'cy', 'br', 'uz', 'gu', 'an',
              'bar', 'ne', 'si', 'lb', 'jv', 'zh-min-nan', 'war', 'sw', 'ga', 'ku', 'ckb', 'oc', 'nds', 'yi', 'ia',
              'fy', 'tt', 'scn', 'pa', 'gan', 'am', 'lmo', 'km', 'tg', 'sa', 'ba', 'azb', 'io', 'as', 'vo', 'ky', 'pnb',
              'vec', 'so', 'cv', 'or', 'hak', 'pdc', 'hif', 'ce', 'bh', 'mg', 'su', 'mzn', 'ht', 'nap', 'qu', 'ps',
              'fo', 'li', 'se', 'bo', 'gd', 'pms', 'nds-nl', 'new', 'bat-smg', 'vls', 'yo', 'rue', 'diq', 'ace', 'tk',
              'bpy', 'dv', 'hsb', 'eml', 'cu', 'os', 'wa', 'sah', 'ksh', 'sc', 'chr', 'szl', 'nah', 'mt', 'lad', 'co',
              'pam', 'ug', 'bcl', 'cdo', 'arc', 'rm', 'gv', 'got', 'frr', 'dsb', 'ab', 'crh', 'xmf', 'zu', 'iu', 'rmy',
              'cr', 'ie', 'ilo', 'gn', 'ext', 'mi', 'ha', 'csb', 'ay', 'pcd', 'sd', 'map-bms', 'min', 'lo', 'jbo', 'nv',
              'sn', 'haw', 'frp', 'vep', 'ch', 'glk', 'lij', 'wo', 'udm', 'cbk-zam', 'kw', 'bxr', 'pap', 'ee', 'fur',
              'av', 'kv', 'roa-rup', 'fiu-vro', 'mhr', 'ig', 'stq', 'bjn', 'nrm', 'mwl', 'bug', 'kl', 'gag', 'tpi',
              'bi', 'zea', 'kab', 'ak', 'ln', 'myv', 'tw', 'xh', 'na', 'mai', 'roa-tara', 'nov', 'rw', 'pfl', 'chy',
              'pih', 'kaa', 'mrj', 'kg', 'bm', 'krc', 'za', 'sm', 'lez', 'pnt', 'xal', 'st', 'om', 'kbd', 'to', 'dz',
              'tn', 'ks', 'tet', 'ts', 'rn', 'ny', 'mdf', 'gom', 'ti', 'fj', 'lfn', 'koi', 'lbe', 'ik', 'tyv', 'ki',
              'ff', 'pag', 'ss', 'tum', 'srn', 'lg', 'ty', 've', 'jam', 'ltg', 'pi', 'hyw', 'sg', 'kr', 'olo', 'nso',
              'ady', 'din', 'lrc', 'dty', 'tcy', 'sat', 'aa', 'hz', 'ary', 'ban', 'kbp', 'atj', 'gor', 'shn', 'inh',
              'ng', 'mus', 'mh', 'nqo', 'ii', 'mnw', 'avk', 'szy', 'cho', 'gcr', 'ho', 'kj', 'smn', 'awa', 'lld', 'mad',
              'alt', 'mni', 'dag', 'skr', 'nia', 'trv', 'tay', 'shi']


@lang.command()
@click.argument('indir', type=click.Path(exists=True, file_okay=False))
@click.option('-s', '--split', help='Which input split to use', default='train',
              type=click.Choice(['train', 'test', 'val']), show_default=True)
@click.option('-f', '--out-format', help='Output format (raw vectors or C code)', default='raw',
              type=click.Choice(['raw', 'c']), show_default=True)
@click.option('-l', '--vector-size', help='Output vector size', default=256, type=int, show_default=True)
def train_vectors(indir, split, out_format, vector_size):
    """
    Train and print vectors for fast language detection.

    Expects the following directory structure:

    ..
        indir/
            <LANGCODE 1>/
                test.txt
                train.txt
                val.txt
            <LANGCODE 2>/
                test.txt
                train.txt
                val.txt
            ...
    """

    # Sort descending by amount of data for biasing towards common languages
    langs = os.listdir(indir)
    langs.sort(key=lambda l: _WIKI_BIAS.index(l) if l in _WIKI_BIAS else len(_WIKI_BIAS) + langs.index(l))

    if out_format == 'c':
        click.echo(f'''/*
    Resiliparse fast language detection profiles.
    Generated automatically, do not modify.
*/

#ifndef RESILIPARSE_LANG_PROFILES_H
#define RESILIPARSE_LANG_PROFILES_H

#include <stdint.h>

#define LANG_VEC_SIZE {vector_size}
typedef const uint8_t lang_vec_t[LANG_VEC_SIZE];

typedef struct lang {{
    const char* lang;
    const lang_vec_t vec;
}} lang_t;

/* Sorted by number of Wikipedia users */
static const lang_t LANGS[] = {{''', nl=False)
    else:
        click.echo('# (lang, vec)')

    for i, l in enumerate(langs):
        vec = rlang.train_language_examples(open(os.path.join(indir, l, split + '.txt'), 'r'), vector_size)
        if out_format == 'c':
            if i > 0:
                click.echo(',', nl=False)
            click.echo(f'\n    {{"{l}", {{{", ".join(str(i) for i in vec)}}}}}', nl=False)
        else:
            click.echo((l, vec))

    if out_format == 'c':
        click.echo('\n};\n')
        click.echo('#define N_LANGS sizeof(LANGS) / sizeof(lang_t)\n')
        click.echo('#endif  // RESILIPARSE_LANG_PROFILES_H')


@lang.command()
@click.argument('indir', type=click.Path(exists=True, file_okay=False))
@click.option('-s', '--split', help='Which input split to use', type=click.Choice(['val', 'test']),
              default='val', show_default=True)
@click.option('-l', '--langs', help='Restrict languages to this comma-separated list')
@click.option('-c', '--cutoff', type=int, help='Prediction cutoff', default=1200, show_default=True)
@click.option('-t', '--truncate', type=int, help='Truncate examples to this length')
@click.option('-f', '--fasttext-model', help='Use the specified FastText model for samples above cutoff',
              type=click.Path(exists=True, dir_okay=False))
@click.option('--sort-lang', is_flag=True, help='Sort by language instead of F1')
@click.option('--print-cm', is_flag=True, help='Print confusion matrix (may be very big)')
def evaluate(indir, split, langs, truncate, cutoff, sort_lang, print_cm, fasttext_model):
    """
    Evaluate language prediction performance.
    """
    if langs is not None:
        langs = {l.strip() for l in langs.split(',')}
    in_langs = sorted([l for l in os.listdir(indir) if langs is None or l in langs])

    if fasttext_model:
        try:
            import fasttext
        except ModuleNotFoundError:
            click.echo('FastText needs to be installed if --fasttext-model is set.', err=True)
            click.echo('Run "pip install fasttext" to install it.', err=True)
            return

        # FastText prints useless warnings after loading a model, so silence its print method to make it shut up
        # See: https://github.com/facebookresearch/fastText/issues/909
        fasttext.FastText.eprint = lambda x: None
        fasttext_model = fasttext.load_model(fasttext_model)

    recall_matrix = defaultdict(list)
    precision_matrix = defaultdict(list)
    confusion_matrix = defaultdict(lambda: defaultdict(int))

    for lang in tqdm(in_langs, desc='Evaluating languages', unit='language', leave=False):
        for line in tqdm(open(os.path.join(indir, lang, split + '.txt'), 'r'),
                         desc=f'Predicting examples for {lang}', unit=' examples', leave=False):

            if truncate:
                line = line[:truncate]

            plang = rlang.detect_fast(line, cutoff=cutoff, langs=langs)[0]
            if plang == 'unknown':
                if fasttext_model:
                    ft_pred = fasttext_model.predict(line.replace('\n', ' '))
                    plang = ft_pred[0][0].replace('__label__', '') if ft_pred else '-'
                else:
                    plang = '-'
            recall_matrix[lang].append(plang == lang)
            precision_matrix[plang].append(plang == lang)
            confusion_matrix[lang][plang] += 1

    acc = []
    results = []
    sum_examples = 0
    for lang in in_langs:
        precision = sum(precision_matrix[lang]) / max(1, len(precision_matrix[lang]))
        recall = sum(recall_matrix[lang]) / max(1, len(recall_matrix[lang]))
        if precision + recall == 0:
            f1 = 0.0
        else:
            f1 = 2.0 * (precision * recall) / (precision + recall)
        n_ex = len(recall_matrix[lang])
        results.append((lang, precision, recall, f1, n_ex))
        acc.append(recall * n_ex)
        sum_examples += n_ex

    click.echo('Lang, Precision, Recall, F1, Top Confusions, Num Examples')
    if not sort_lang:
        results.sort(key=lambda x: x[3], reverse=True)
    for r in results:
        click.echo(f'{r[0]}, {r[1]:.2f}, {r[2]:.2f}, {r[3]:.2f}, {r[4]}')

    acc = fsum(acc) / sum_examples
    click.echo(f'\nAccuracy: {acc:.2f}')

    if print_cm:
        label_width = max(len(l) for l in in_langs)
        col_width = max(max(max(len(str(l2)) for l2 in l1.values()) for l1 in confusion_matrix.values()), label_width) + 2

        click.echo(f'\nConfusion matrix:\n' + (' ' * label_width), nl=False)
        for l in in_langs:
            click.echo(f'{l:>{col_width}}', nl=False)

        click.echo()
        for l1 in in_langs:
            click.echo(l1, nl=False)
            for l2 in in_langs:
                click.echo(f'{confusion_matrix[l1][l2]:>{col_width}}', nl=False)
            click.echo()


@lang.command()
@click.argument('infile', type=click.Path(exists=True, dir_okay=False))
@click.option('-r', '--rounds', help='Number of rounds to benchmark', type=int, default=10000, show_default=True)
@click.option('-f', '--fasttext-model', help='FastText model to benchmark', type=click.Path(exists=True, dir_okay=False))
def benchmark(infile, rounds, fasttext_model):
    """
    Benchmark Resiliparse against FastText and Langid.

    Either package must be installed for this comparison. Install Resiliparse with the
    ``cli-benchmark`` flag to install all optional third-party dependencies automatically.
    """

    if fasttext_model:
        try:
            import fasttext
        except ModuleNotFoundError:
            click.echo('FastText needs to be installed if --fasttext-model is set.', err=True)
            click.echo('Run "pip install fasttext" to install it.', err=True)
            return

        # FastText prints useless warnings after loading a model, so silence its print method to make it shut up
        # See: https://github.com/facebookresearch/fastText/issues/909
        fasttext.FastText.eprint = lambda x: None
        fasttext_model = fasttext.load_model(fasttext_model)
    else:
        click.echo('Skipping FastText benchmark, since no model has been specified.', err=True)

    bench_langid = True
    try:
        import langid
    except ModuleNotFoundError:
        click.echo('Skipping langid benchmark, since it is not installed.', err=True)
        click.echo('Run "pip install langid" to install it.', err=True)
        bench_langid = False

    in_data = bytes_to_str(open(infile, 'rb').read().replace(b'\n', b' '))
    click.echo(f'Benchmarking language detectors ({rounds:,} rounds):')

    start = time.monotonic()
    for _ in tqdm(range(rounds), desc='Benchmarking Resiliparse', unit=' rounds', leave=False, miniters=0.3):
        rlang.detect_fast(in_data)
    print(f'Resiliparse: {time.monotonic() - start:.1f}s')

    if fasttext_model:
        start = time.monotonic()
        for _ in tqdm(range(rounds), desc='Benchmarking FastText', unit=' rounds', leave=False, miniters=0.3):
            fasttext_model.predict(in_data)
        print(f'FastText:    {time.monotonic() - start:.1f}s')

    if bench_langid:
        start = time.monotonic()
        for _ in tqdm(range(rounds), desc='Benchmarking Langid', unit=' rounds', leave=False, miniters=0.3):
            langid.classify(in_data)
        print(f'Langid:      {time.monotonic() - start:.1f}s')


if __name__ == '__main__':
    main()



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

from cpython cimport PyObject, PyThreadState_SetAsyncExc
cimport cython
from libc.stdint cimport uint64_t
from libc.signal cimport SIGINT, SIGTERM, SIGKILL
from libcpp.string cimport string, to_string

import inspect
import platform
import threading
import warnings

from resiliparse_inc.cstdlib cimport strtol
from resiliparse_inc.pthread cimport pthread_kill, pthread_t, pthread_self
from resiliparse_inc.stdio cimport FILE, fclose, feof, fgets, fopen
from resiliparse_inc.time cimport timespec, clock_gettime, CLOCK_MONOTONIC
from resiliparse_inc.unistd cimport getpagesize, getpid, usleep


class ResiliparseGuardException(BaseException):
    """Resiliparse guard base exception."""


class ExecutionTimeout(ResiliparseGuardException):
    """Execution timeout exception."""


class MemoryLimitExceeded(ResiliparseGuardException):
    """Memory limit exceeded exception."""


cdef size_t MAX_SIZE_T = <size_t>-1


cdef inline uint64_t time_millis() nogil:
    cdef timespec t
    clock_gettime(CLOCK_MONOTONIC, &t)
    return t.tv_sec * 1000u + t.tv_nsec / <long>1e6


__GUARD_CTX_ACTIVE = set()


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class _ResiliparseGuard:
    """Resiliparse context guard base class."""

    def __cinit__(self, *args, **kwargs):
        self.gctx.epoch_counter.store(0)
        self.gctx.ended.store(False)
        self.guard_thread = None

    def __dealloc__(self):
        self.finish()

    cdef inline bint setup(self) except 0:
        global __GUARD_CTX_ACTIVE
        if self.__class__.__name__ in __GUARD_CTX_ACTIVE:
            raise RuntimeError('Guard contexts of the same type cannot be nested.')
        __GUARD_CTX_ACTIVE.add(self.__class__.__name__)
        self.gctx.epoch_counter.store(0)
        self.gctx.ended.store(False)
        return True

    cdef inline bint start_guard_thread(self, func, args=()) except 0:
        self.guard_thread = threading.Thread(target=func, args=args, daemon=True)
        self.guard_thread.start()
        return True

    cdef inline bint finish(self) except 0:
        self.gctx.ended.store(True)
        if self.guard_thread is not None:
            self.guard_thread.join()
            self.guard_thread = None

        global __GUARD_CTX_ACTIVE
        if __GUARD_CTX_ACTIVE is not None and self.__class__.__name__ in __GUARD_CTX_ACTIVE:
            __GUARD_CTX_ACTIVE.remove(self.__class__.__name__)

        return True

    def __call__(self, func):
        def guard_wrapper(*args, **kwargs):
            self.setup()
            self.exec_before()
            try:
                return func(*args, **kwargs)
            finally:
                self.exec_after()
                self.finish()

        # Retain self, but do not bind via __get__() or else func will belong to this class
        guard_wrapper._guard_self = self

        # Decorate with public methods of guard instance for convenience
        for attr in dir(self):
            if not attr.startswith('_'):
                setattr(guard_wrapper, attr, getattr(self, attr))

        return guard_wrapper

    def __enter__(self):
        self.setup()
        self.exec_before()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exec_after()
        self.finish()

    cdef void exec_before(self) except *:
        pass

    cdef void exec_after(self) except *:
        pass

    cdef type get_exception_type(self):
        """Interrupt exception type to send"""
        pass

    cdef void send_interrupt(self, unsigned char escalation_level, pthread_t target_thread) nogil:
        if self.gctx.ended.load():
            return

        if escalation_level == 0:
            if self.interrupt_type == exception or self.interrupt_type == exception_then_signal:
                with gil:
                    self.exc_type = self.get_exception_type()
                    PyThreadState_SetAsyncExc(<unsigned long>target_thread, <PyObject*>self.exc_type)
            elif self.interrupt_type == signal:
                pthread_kill(target_thread, SIGINT)

        elif escalation_level == 1:
            if self.interrupt_type == signal:
                pthread_kill(target_thread, SIGTERM)
            elif self.interrupt_type == exception_then_signal:
                pthread_kill(target_thread, SIGINT)
            elif self.interrupt_type == exception:
                with gil:
                    self.exc_type = self.get_exception_type()
                    PyThreadState_SetAsyncExc(<unsigned long>target_thread, <PyObject*>self.exc_type)

        elif escalation_level == 2:
            if self.interrupt_type != exception and self.send_kill:
                pthread_kill(target_thread, SIGKILL)
            elif self.interrupt_type != exception:
                pthread_kill(target_thread, SIGTERM)
            elif self.interrupt_type == exception:
                with gil:
                    self.exc_type = self.get_exception_type()
                    PyThreadState_SetAsyncExc(<unsigned long>target_thread, <PyObject*>self.exc_type)
            with gil:
                warnings.warn('ERROR: Guarded thread did not respond to TERM signal.', RuntimeWarning)


@cython.auto_pickle(False)
cdef class TimeGuard(_ResiliparseGuard):
    """
    Decorator and context manager for guarding the execution time of a running task.

    Use the :func:`time_guard` factory function for instantiation.
    """

    # noinspection PyMethodOverriding
    def __cinit__(self, size_t timeout, size_t grace_period, InterruptType interrupt_type, bint send_kill,
                  size_t check_interval):
        self.timeout = timeout
        self.grace_period = max(10u, grace_period)
        self.interrupt_type = interrupt_type
        self.send_kill = send_kill
        self.check_interval = check_interval

    cdef type get_exception_type(self):
        return ExecutionTimeout

    cdef void exec_before(self) except *:
        cdef pthread_t main_thread = pthread_self()

        def _thread_exec():
            cdef uint64_t last_epoch = 0
            cdef uint64_t now = time_millis()
            cdef uint64_t start = now
            cdef unsigned char signals_sent = 0

            with nogil:
                while True:
                    usleep(self.check_interval * 1000)

                    if self.gctx.ended.load():
                        break

                    now = time_millis()

                    if self.gctx.epoch_counter.load() > last_epoch:
                        start = now
                        last_epoch = self.gctx.epoch_counter.load()
                        signals_sent = 0

                    # Exceeded, but within grace period
                    if (self.timeout == 0 or now - start >= self.timeout) and signals_sent == 0:
                        signals_sent = 1
                        self.send_interrupt(0, main_thread)

                    # Grace period exceeded
                    elif now - start >= (self.timeout + self.grace_period) and signals_sent == 1:
                        signals_sent = 2
                        self.send_interrupt(1, main_thread)

                    # If process still hasn't reacted, send SIGTERM/SIGKILL and then exit
                    elif now - start >= (self.timeout + self.grace_period * 2) and signals_sent == 2:
                        signals_sent = 3
                        self.send_interrupt(2, main_thread)
                        break

        self.start_guard_thread(_thread_exec)

    cpdef void progress(self):
        """
        progress(self)
        
        Increment epoch counter to indicate progress and reset the guard timeout.
        This method is thread-safe.
        """
        self.gctx.epoch_counter.store(time_millis())


def time_guard(timeout=60, timeout_ms=None, grace_period=15, grace_period_ms=None,
               InterruptType interrupt_type=exception_then_signal, bint send_kill=False,
               size_t check_interval=500) -> TimeGuard:
    """
    time_guard(timeout=60, timeout_ms=None, grace_period=15, grace_period_ms=None, \
               interrupt_type=exception_then_signal, send_kill=False, check_interval=500)

    Create a :class:`TimeGuard` instance that can be used as a decorator or context manager
    for guarding the execution time of a running task.

    If a the guarded context runs longer than the pre-defined timeout, the guard will send
    an interrupt to the running thread. The timeout can be reset at any time by proactively
    reporting progress to the guard instance. This can be done by calling
    :meth:`TimeGuard.progress` on the guard instance or the convenience function
    :func:`progress` (if the context is in the global scope).

    There are two interrupt mechanisms: throwing an asynchronous exception and sending
    a UNIX signal. The exception mechanism is the most gentle method of the two, but
    may be unreliable if execution is blocking outside the Python program flow (e.g.,
    in a native C extension or in a syscall). The signal method is more reliable
    in this regard, but does not work if the guarded thread is not the interpreter main
    thread, since only the main thread can receive and handle signals.

    The Interrupt behaviour can be configured with the ``interrupt_type`` parameter, which
    accepts an enum value of type :class:`InterruptType`:

    If ``interrupt_type`` is :attr:`~InterruptType.exception`, an :exc:`ExecutionTimeout`
    exception will be sent to the running thread after `timeout` seconds. If the thread
    does not react, the exception will be thrown once more after `grace_period` seconds.

    If ``interrupt_type`` is :attr:`~InterruptType.signal`, first a ``SIGINT`` will be sent to the
    current thread (which will trigger a :exc:`KeyboardInterrupt` exception, but can
    also be handled with a custom ``signal`` handler). If the thread does not react, a less
    friendly ``SIGTERM`` will be sent after ``grace_period`` seconds. A third and final
    attempt of a ``SIGTERM`` will be sent after another ``grace_period`` seconds.

    If ``interrupt_type`` is :attr:`~InterruptType.exception_then_signal` (the default), the
    first attempt will be an exception and after the grace period, the guard will start
    sending signals.

    With ``send_kill`` set to ``True``, the third and final attempt will be a ``SIGKILL`` instead
    of a ``SIGTERM``. This will kill the entire interpreter (even if the guarded thread is not
    the main thread), so you will need an external facility to restart it.

    :param timeout: max execution time in seconds before invoking interrupt
    :type timeout: int
    :param timeout_ms: max execution time in milliseconds (use instead of ``timeout`` if higher resolution needed)
    :type timeout_ms: int
    :param grace_period: grace period in seconds after which to send another interrupt
    :type grace_period: int
    :param grace_period_ms: grace period in milliseconds (use instead of ``grace_period`` if higher resolution needed)
    :type grace_period_ms: int
    :param interrupt_type: type of interrupt
    :type interrupt_type: InterruptType
    :param send_kill: send ``SIGKILL`` as third attempt instead of ``SIGTERM`` (ignored if
                      ``interrupt_type`` is :attr:`~InterruptType.exception`)
    :type send_kill: bool
    :param check_interval: interval in milliseconds between execution time checks
    :type check_interval: int
    :rtype: TimeGuard
    """
    if timeout is timeout_ms is None:
        raise ValueError('No timeout set')

    timeout = timeout * 1000 if timeout_ms is None else timeout_ms
    grace_period = grace_period * 1000 if grace_period_ms is None else grace_period_ms

    return TimeGuard.__new__(TimeGuard, timeout, grace_period, interrupt_type, send_kill, check_interval)


cpdef progress(ctx=None):
    """
    progress(ctx=None)
    
    Increment :class:`TimeGuard` epoch counter to indicate progress and reset the guard timeout
    for the active guard context surrounding the caller.

    If ``ctx`` ist ``None``, the last valid guard context from the global namespace on
    the call stack will be used. If the guard context does not live in the module's
    global namespace, this auto-detection will fail and the caller has to be supplied
    explicitly.

    If no valid guard context can be determined, the progress report will fail and a
    :class:`RuntimeError` will be raised.

    :param ctx: active guard context (will use last global context from stack if unset)
    :raise RuntimeError: if no valid :class:`TimeGuard` found
    """
    if ctx is None:
        for i in range(len(inspect.stack())):
            frame_info = inspect.stack()[i]
            ctx = frame_info[0].f_globals.get(frame_info[3])
            if isinstance(ctx, TimeGuard) or isinstance(getattr(ctx, '_guard_self', None), TimeGuard):
                break

    if isinstance(ctx, TimeGuard):
        (<TimeGuard>ctx).progress()
        return

    if not isinstance(getattr(ctx, '_guard_self', None), TimeGuard):
        raise RuntimeError('No initialized time guard context.')

    # noinspection PyProtectedMember, PyUnresolvedReferences
    (<TimeGuard>ctx._guard_self).progress()


def progress_loop(it, ctx=None):
    """
    progress_loop(it, ctx=None)

    Wraps an iterator into a pass-through iterator that reports progress
    to an active :class:`resiliparse.process_guard.TimeGuard` context guard after each iteration.

    :param it: original iterator
    :type it: t.Iterable[t.Any]
    :param ctx: active guard context (will use last global context from stack if unset)
    :return: wrapped iterator
    :rtype: t.Iterable[t.Any]
    """
    for i in it:
        yield i
        progress(ctx)


@cython.auto_pickle(False)
cdef class MemGuard(_ResiliparseGuard):
    """
    Decorator and context manager for enforcing memory limits on a running task.

    Use the :func:`mem_guard` factory function for instantiation.
    """

    # noinspection PyMethodOverriding
    def __cinit__(self, size_t max_memory, bint absolute, size_t grace_period, size_t secondary_grace_period,
                  InterruptType interrupt_type, bint send_kill, size_t check_interval):
        self.max_memory = max_memory
        self.absolute = absolute
        self.grace_period = grace_period
        self.secondary_grace_period = max(10u, secondary_grace_period)
        self.interrupt_type = interrupt_type
        self.send_kill = send_kill
        self.check_interval = check_interval

    # cdef size_t _get_rss_posix(self) nogil:
    #     cdef string cmd = string(<char*>b'ps -p ').append(to_string(getpid())).append(<char*>b' -o rss=')
    #     cdef string buffer = string(64, <char>0)
    #     cdef string out
    #     cdef FILE* fp = popen(cmd.c_str(), <char*>b'r')
    #     if fp == NULL:
    #         return 0
    #     while not feof(fp):
    #         if fgets(buffer.data(), 64, fp) != NULL:
    #             out.append(buffer.data())
    #     pclose(fp)
    #     return strtol(out.c_str(), NULL, 10)

    cdef inline size_t _get_rss(self) nogil:
        cdef string proc_file = string(<char*>b'/proc/').append(to_string(getpid())).append(<char*>b'/statm')
        cdef string buffer = string(64, <char>0)
        cdef string statm
        cdef FILE* fp = fopen(proc_file.c_str(), <char*>b'r')
        if fp == NULL:
            return 0
        while not feof(fp):
            if fgets(buffer.data(), 64, fp) != NULL:
                statm.append(buffer.data())
        fclose(fp)

        statm = statm.substr(statm.find(<char*>b' ') + 1)   # VmSize (skip)
        statm = statm.substr(0, statm.find(<char*>b' '))    # VmRSS
        return strtol(statm.c_str(), NULL, 10) * getpagesize() // 1024u

    cdef type get_exception_type(self):
        return MemoryLimitExceeded

    cdef void exec_before(self) except *:
        if platform.system() != 'Linux':
            raise RuntimeError(f'Unsupported platform: {platform.system()}')

        cdef pthread_t main_thread = pthread_self()

        cdef size_t max_mem = self.max_memory
        if not self.absolute:
            max_mem += self._get_rss()

        def _thread_exec():
            cdef uint64_t grace_start = 0
            cdef uint64_t now = 0
            cdef unsigned char signals_sent = 0
            cdef size_t rss = 0

            with nogil:
                while True:
                    rss = self._get_rss()

                    if rss > max_mem:
                        # Memory usage above limit, start grace period
                        now = time_millis()

                        if grace_start == 0:
                            grace_start = now
                            signals_sent = 0

                        # Grace period exceeded
                        if (self.grace_period == 0 or now - grace_start >= self.grace_period) and signals_sent == 0:
                            signals_sent = 1
                            self.send_interrupt(0, main_thread)

                        # Secondary grace period exceeded
                        elif now - grace_start >= self.grace_period + self.secondary_grace_period \
                                and signals_sent == 1:
                            signals_sent = 2
                            self.send_interrupt(1, main_thread)

                        # If process still hasn't reacted, send SIGTERM/SIGKILL and then exit
                        elif now - grace_start >= self.grace_period + self.secondary_grace_period * 2 \
                                and signals_sent == 2:
                            signals_sent = 3
                            self.send_interrupt(2, main_thread)
                            break

                    elif rss < max_mem and grace_start != 0:
                        # Memory usage dropped, reset grace period
                        grace_start = 0
                        signals_sent = 0

                    usleep(self.check_interval * 1000)
                    if self.gctx.ended.load():
                        break

        self.start_guard_thread(_thread_exec)


def mem_guard(size_t max_memory, bint absolute=True, grace_period=0, grace_period_ms=None,
              secondary_grace_period=5, secondary_grace_period_ms=None,
              InterruptType interrupt_type=exception_then_signal, bint send_kill=False,
              size_t check_interval=500) -> MemGuard:
    """
    mem_guard(max_memory, absolute=True, grace_period=0, grace_period_ms=0, secondary_grace_period=5, \
              secondary_grace_period_ms=None, interrupt_type=exception_then_signal, send_kill=False, \
              check_interval=500)

    Create a :class:`MemGuard` instance that can be used as a decorator or context manager
    for enforcing memory limits on a running task.

    :class:`MemGuard` guards a function or other context to stay within pre-defined memory
    bounds. If the running Python process exceeds these bounds while the guard context
    is active, an exception or signal will be sent to the executing thread.

    If the thread does not react to this exception, the same escalation procedure will
    kick in as known from :class:`TimeGuard`. In order for :class:`MemGuard` to tolerate
    short spikes above the memory limit, set the ``grace_period`` parameter to a
    positive non-zero value. If memory usage exceeds the limit, a timer will start
    that expires after ``grace_period`` seconds and triggers the interrupt procedure.
    If memory usage falls below the threshold during the grace period, the timer is reset.

    :class:`MemGuard` provides the same parameters as :class:`TimeGuard` for controlling
    the interrupt escalation behaviour, but the time interval before triggering the
    next escalation level is independent of the grace period and defaults to five
    seconds to give the application sufficient time to react and deallocate excess memory.
    This secondary grace period can be configured with the ``secondary_grace_period``
    parameter and must be at least one second.

    .. warning::

        :class:`MemGuard` is supported only on Linux. You can decorate functions with it
        on other platforms, but calling them will raise an error.

    :param max_memory: max allowed memory in KiB since context creation before interrupt will be sent
    :type max_memory: int
    :param absolute: whether ``max_memory`` is an absolute limit for the process or a relative growth limit
    :type absolute: bool
    :param grace_period: grace period in seconds before sending an interrupt after exceeding ``max_memory``
    :type grace_period: int
    :param grace_period_ms: grace period in milliseconds (use instead of ``grace_period`` if higher resolution needed)
    :type grace_period_ms: int
    :param secondary_grace_period: time to wait after ``grace_period`` before triggering next escalation level
    :type secondary_grace_period: int
    :param secondary_grace_period_ms: secondary grace period in milliseconds (use instead of ``secondary_grace_period``
                                      if higher resolution needed)
    :type secondary_grace_period_ms: int
    :param interrupt_type: type of interrupt
    :type interrupt_type: InterruptType
    :param send_kill: send ``SIGKILL`` as third attempt instead of ``SIGTERM`` (ignored if
                     ``interrupt_type`` is :attr:`~InterruptType.exception`)
    :type send_kill: bool
    :param check_interval: interval in milliseconds between memory consumption checks
    :type check_interval: int
    :rtype: MemGuard
    """
    grace_period = grace_period * 1000 if grace_period_ms is None else grace_period_ms
    secondary_grace_period = secondary_grace_period * 1000 \
        if secondary_grace_period_ms is None else secondary_grace_period_ms

    return MemGuard.__new__(MemGuard, max_memory, absolute, grace_period, secondary_grace_period,
                            interrupt_type, send_kill, check_interval)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from apache_beam.coders import *
from apache_beam.coders import StrUtf8Coder as _StrUtf8Coder
from resiliparse.parse.encoding import bytes_to_str


class StrUtf8Coder(_StrUtf8Coder):
    """
    More resilient version of :class:`apache_beam.coders.StrUtf8Coder`, which can handle encoding errors.

    Uses the :class:`~resiliparse.parse.encoding.bytes_to_str` encoding helpers for encoding and decoding text.
    """

    def decode(self, value):
        return bytes_to_str(value, encoding='utf-8')



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import apache_beam as beam
from apache_beam.io import fileio as beam_fio


class MatchFiles(beam.PTransform):
    """
    Match a file pattern using :meth:`apache_beam.io.filesystems.FileSystems.match`.

    Unlike the original Beam implementation, this file matcher enforces a fusion break
    by reshuffling the matched file names. This circumvents limitations in certain
    Beam runners that do not automatically distribute splits, such as the FlinkRunner.

    :param file_pattern: file glob
    :param empty_match_treatment: what to do with empty glob matches
    :param shuffle: shuffle matches to break fusion (setting this to ``False`` effectively
                    falls back to the original Beam implementation)
    """

    def __init__(self,
                 file_pattern: str,
                 empty_match_treatment: beam_fio.EmptyMatchTreatment = beam_fio.EmptyMatchTreatment.ALLOW_IF_WILDCARD,
                 shuffle: bool = True):
        super().__init__()
        self._transform = beam_fio.MatchFiles(file_pattern, empty_match_treatment)
        if shuffle:
            self._transform |= beam.Reshuffle()

    def expand(self, pcoll):
        return pcoll | self._transform



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging

import apache_beam as beam
import apache_beam.typehints as t
from apache_beam.io.aws.s3io import S3IO, S3Downloader
from apache_beam.io.aws.clients.s3 import boto3_client
from apache_beam.io.filesystem import CompressionTypes
from apache_beam.io.filesystemio import DownloaderStream
from apache_beam.io.filesystems import FileSystems
from apache_beam.io.restriction_trackers import OffsetRange, OffsetRestrictionTracker
from apache_beam.options.value_provider import RuntimeValueProvider
import apache_beam.transforms.window as window
import boto3
import botocore.client as boto_client

from fastwarc.warc import ArchiveIterator
from resiliparse.beam.fileio import MatchFiles
from resiliparse.itertools import warc_retry

logger = logging.getLogger()


class ReadWarcs(beam.PTransform):
    """
    Read WARC records from files matching a glob pattern.

    :param file_pattern: input file glob pattern
    :param warc_args: arguments to pass to :class:`fastwarc.warc.ArchiveIterator`
    :param with_filename: keep the input filename as a key (otherwise return only the record)
    :param freeze: freeze returned records (required if the records are not consumed immediately)
    :param always_keep_meta: always return record metadata, even if they exceed ``max_content_length``
                             (from ``warc_args``), but strip them of their payload
    :param shuffle_files: shuffle matched file names (useful if Beam runner does not support automatic
                          shuffling of input source splits)
    """

    def __init__(self,
                 file_pattern: str,
                 warc_args: t.Dict[str, t.Any] = None,
                 with_filename: bool = True,
                 freeze: bool = True,
                 always_keep_meta: bool = False,
                 shuffle_files: bool = False):
        super().__init__()
        self._transform = MatchFiles(file_pattern, shuffle=shuffle_files)
        self._transform |= beam.ParDo(_ReadWarc(warc_args=warc_args,
                                                with_filename=with_filename,
                                                freeze=freeze,
                                                always_keep_meta=always_keep_meta))

    def expand(self, pcoll):
        return pcoll | self._transform


class ReadAllWarcs(beam.PTransform):
    """
    Read WARC records from a given :class:`~apache_beam.pvalue.PCollection` of
    :class:`~apache_beam.io.filesystem.FileMetadata` objects.

    :param warc_args: arguments to pass to :class:`fastwarc.warc.ArchiveIterator`
    :param with_filename: keep the input filename as a key (otherwise return only the record)
    :param freeze: freeze returned records (required if returned records are not consumed immediately)
    :param always_keep_meta: always return record metadata, even if they exceed ``max_content_length``
                             (from ``warc_args``), but strip them of their payload
    """
    def __init__(self,
                 warc_args: t.Dict[str, t.Any] = None,
                 with_filename: bool = True,
                 freeze: bool = True,
                 always_keep_meta: bool = False):
        super().__init__()
        self._transform = beam.ParDo(_ReadWarc(warc_args=warc_args,
                                               with_filename=with_filename,
                                               freeze=freeze,
                                               always_keep_meta=always_keep_meta))

    def expand(self, pcoll):
        return pcoll | self._transform


class _WarcRestrictionProvider(beam.transforms.core.RestrictionProvider):
    def initial_restriction(self, file_meta):
        return OffsetRange(0, file_meta.size_in_bytes)

    def create_tracker(self, restriction):
        return OffsetRestrictionTracker(restriction)

    def restriction_size(self, file_meta, restriction):
        return min(file_meta.size_in_bytes, restriction.stop - restriction.start)


# noinspection PyAbstractClass
class _ReadWarc(beam.DoFn):
    """
    WARC file input source.
    """

    def __init__(self, warc_args, with_filename, freeze, always_keep_meta):
        super().__init__()
        self.warc_args = warc_args or {}
        self.with_filename = with_filename
        self.freeze = freeze
        self.max_content_length = None
        if always_keep_meta and 'max_content_length' in self.warc_args:
            self.max_content_length = self.warc_args['max_content_length']
            del self.warc_args['max_content_length']

    # noinspection PyMethodOverriding
    def process(self, file_meta, tracker=beam.DoFn.RestrictionParam(_WarcRestrictionProvider())):
        restriction = tracker.current_restriction()

        def stream_factory():
            return self._open_file(file_meta.path)

        record = None
        stream = None
        try:
            logger.info('Starting WARC file %s', file_meta.path)
            stream = stream_factory()
            for record in warc_retry(ArchiveIterator(stream, **self.warc_args), stream_factory):
                logger.debug('Reading WARC record %s', record.record_id)
                if not tracker.try_claim(record.stream_pos):
                    break

                if self.max_content_length is not None and record.content_length > self.max_content_length:
                    # Max length exceeded, but we still want to return a metadata record
                    logger.debug("Stripping long record %s (%i bytes) of its payload.",
                                 record.record_id, record.content_length)
                    record.reader.consume()

                if self.freeze:
                    record.freeze()

                if self.with_filename:
                    yield window.TimestampedValue((file_meta.path, record), record.record_date.timestamp())
                else:
                    yield window.TimestampedValue(record, record.record_date.timestamp())
            else:
                tracker.try_claim(restriction.stop)
            logger.info('Completed WARC file %s', file_meta.path)
        except Exception as e:
            if record:
                logger.error('WARC reader failed in %s past record %s (pos: %i).',
                             file_meta.path, record.record_id, record.stream_pos)
            else:
                logger.error('WARC reader failed in %s', file_meta.path)
            logger.exception(e)
        finally:
            if stream and not stream.closed:
                stream.close()

    def _open_file(self, file_name):
        """Get input file stream."""
        if file_name.startswith('s3://'):
            stream = self._open_s3_stream(file_name)
        else:
            stream = FileSystems.open(file_name, 'application/octet-stream',
                                      compression_type=CompressionTypes.UNCOMPRESSED)

        return stream

    # noinspection PyProtectedMember
    def _open_s3_stream(self, file_name, buffer_size=65536):
        """Open S3 streams with custom Boto3 client."""

        options = FileSystems._pipeline_options or RuntimeValueProvider.runtime_options
        s3_client = _Boto3Client(options=options)
        s3io = S3IO(client=s3_client, options=options)

        downloader = S3Downloader(s3io.client, file_name, buffer_size=buffer_size)
        return io.BufferedReader(DownloaderStream(downloader, mode='rb'), buffer_size=buffer_size)


class _Boto3Client(boto3_client.Client):    # pragma: no cover
    """Boto3 client with custom settings."""

    # noinspection PyMissingConstructor
    def __init__(self, options, connect_timeout=60, read_timeout=240):
        super().__init__(options)

        options = options.get_all_options()
        session = boto3.session.Session()
        self.client = session.client(
            service_name='s3',
            region_name=options.get('s3_region_name'),
            api_version=options.get('s3_api_version'),
            use_ssl=not options.get('s3_disable_ssl', False),
            verify=options.get('s3_verify'),
            endpoint_url=options.get('s3_endpoint_url'),
            aws_access_key_id=options.get('s3_access_key_id'),
            aws_secret_access_key=options.get('s3_secret_access_key'),
            aws_session_token=options.get('s3_session_token'),
            config=boto_client.Config(
                connect_timeout=connect_timeout,
                read_timeout=read_timeout
            )
        )



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import apache_beam as beam
import apache_beam.typehints as t
from apache_beam.io import fileio as beam_fio
from apache_beam.io.filesystems import FileSystems
from apache_beam.io.filesystem import CompressionTypes, FileMetadata, CompressedFile
from apache_beam.io.restriction_trackers import OffsetRange, OffsetRestrictionTracker

from resiliparse.beam import coders
from resiliparse.beam.fileio import MatchFiles


DEFAULT_DESIRED_SPLIT_SIZE = 64 * 1024 * 1024   # 64 MiB
DEFAULT_MIN_SPLIT_SIZE = 1024 * 1024            # 1 MiB


class ReadFromText(beam.PTransform):
    """
    Read lines from text files in parallel.

    Can be used to parallelize the processing of large individual text files with newline-delimited
    records. Unlike :class:`apache_beam.io.textio.ReadFromText`, :class:`ReadFromText` can prevent
    fusion of the file splits by opportunistically generating splits and shuffling them before
    actually reading the file contents. This prevents input bottlenecks on certain runners that do not
    automatically distribute splits for parallel processing such as the FlinkRunner.

    The only supported newline separator at the moment is ``\\n``.

    :param file_pattern: input file glob
    :param empty_match_treatment: what to do with empty glob matches
    :param shuffle_splits: shuffle input splits to break fusion
    :param shuffle_names: shuffle matched file names before deriving splits (adds one more fusion
                          break, should be enabled if ``file_pattern`` matches many files or if
                          input files are compressed and cannot be split)
    :param desired_split_size: desired file split size in bytes
    :param min_split_size: minimum file split size in bytes
    :param compression_type: file compression type, will determine whether individual files are splittable
    :param coder: coder for decoding file contents
    """

    def __init__(self,
                 file_pattern: str,
                 empty_match_treatment: beam_fio.EmptyMatchTreatment = beam_fio.EmptyMatchTreatment.ALLOW_IF_WILDCARD,
                 shuffle_splits: bool = True,
                 shuffle_names: bool = False,
                 desired_split_size: int = DEFAULT_DESIRED_SPLIT_SIZE,
                 min_split_size: int = DEFAULT_MIN_SPLIT_SIZE,
                 compression_type: CompressionTypes = CompressionTypes.AUTO,
                 coder: coders.Coder = coders.StrUtf8Coder()):
        super().__init__()
        self._transform = MatchFiles(file_pattern, empty_match_treatment, shuffle_names)
        self._transform |= beam.ParDo(_GenSplits(desired_split_size, min_split_size, compression_type))
        if shuffle_splits:
            self._transform |= beam.Reshuffle()
        self._transform |= beam.ParDo(_ReadSplits(compression_type, coder))

    def expand(self, pcoll):
        return pcoll | self._transform


class ReadAllFromText(beam.PTransform):
    """
    Read lines from a given :class:`~apache_beam.pvalue.PCollection` of
    :class:`~apache_beam.io.filesystem.FileMetadata` objects in parallel.

    See :class:`ReadFormText` for more information.

    :param shuffle_splits: shuffle input splits to break fusion
    :param shuffle_names: shuffle matched file names before deriving splits (adds one more fusion
                          break, should be enabled if ``file_pattern`` matches many files or if
                          input files are compressed and cannot be split)
    :param desired_split_size: desired file split size in bytes
    :param min_split_size: minimum file split size in bytes
    :param compression_type: file compression type, will determine whether individual files are splittable
    :param coder: coder for decoding file contents
    """

    def __init__(self,
                 shuffle_splits: bool = True,
                 shuffle_names: bool = False,
                 desired_split_size: int = DEFAULT_DESIRED_SPLIT_SIZE,
                 min_split_size: int = DEFAULT_MIN_SPLIT_SIZE,
                 compression_type: CompressionTypes = CompressionTypes.AUTO,
                 coder: coders.Coder = coders.StrUtf8Coder()):
        super().__init__()
        self._transform = beam.ParDo(_GenSplits(desired_split_size, min_split_size, compression_type))
        if shuffle_names:
            self._transform = beam.Reshuffle() | self._transform
        if shuffle_splits:
            self._transform |= beam.Reshuffle()
        self._transform |= beam.ParDo(_ReadSplits(compression_type, coder))

    def expand(self, pcoll):
        return pcoll | self._transform


# noinspection PyAbstractClass
class _GenSplits(beam.DoFn):
    """Create initial splits based on the text file size."""

    def __init__(self, desired_split_size, min_split_size, compression_type):
        super().__init__()
        self.desired_split_size = max(min_split_size, desired_split_size)
        self.min_split_size = min_split_size
        self.compression_type = compression_type

    # noinspection PyMethodOverriding
    def process(self, element: FileMetadata):
        comp_type = self.compression_type
        if self.compression_type == CompressionTypes.AUTO:
            comp_type = CompressionTypes.detect_compression_type(element.path)

        if comp_type != CompressionTypes.UNCOMPRESSED:
            yield element, (0, element.size_in_bytes)
            return

        for start_offset in range(0, element.size_in_bytes, self.desired_split_size):
            end_offset = min(start_offset + self.desired_split_size, element.size_in_bytes)
            if element.size_in_bytes - end_offset < self.min_split_size:
                yield element, (start_offset, element.size_in_bytes)
                break
            yield element, (start_offset, end_offset)


class _FileSplitRestrictionProvider(beam.transforms.core.RestrictionProvider):
    def initial_restriction(self, file_split):
        return OffsetRange(file_split[1][0], file_split[1][1])

    def create_tracker(self, restriction):
        return OffsetRestrictionTracker(restriction)

    def restriction_size(self, file_split, restriction):
        return file_split[1][1] - file_split[1][0]


# noinspection PyAbstractClass
class _ReadSplits(beam.DoFn):
    """Read text lines from splits."""

    def __init__(self, compression_type, coder):
        super().__init__()
        self.compression_type = compression_type
        self.coder = coder

    # noinspection PyMethodOverriding
    def process(self,
                element: t.KV[FileMetadata, t.Tuple[int, int]],
                tracker=beam.DoFn.RestrictionParam(_FileSplitRestrictionProvider())):
        file_meta, _ = element

        restriction = tracker.current_restriction()
        pos = restriction.start

        with FileSystems.open(file_meta.path,
                              mime_type=CompressionTypes.mime_type(self.compression_type),
                              compression_type=self.compression_type) as file:

            if isinstance(file, CompressedFile) and tracker.try_claim(pos):
                # If file is compressed, we need to check the position in the compressed stream
                if pos > 0:
                    file.seek(pos)
                while file._file.tell() <= restriction.stop:
                    line = file.readline()
                    if not line:
                        break
                    yield line[:-1]     # Strip trailing newline character

                tracker.try_claim(file._file.tell())
                return

            if pos > 0:
                # Seek to one character before the starting position and discard the first line.
                # This avoids skipping lines if files are split exactly at newlines.
                file.seek(pos - 1)
                file.readline()

            while tracker.try_claim(file.tell()):
                line = file.readline()
                if not line:
                    raise IOError(f'Unexpected EOF at {file.tell()}')
                line = line[:-1]     # Strip trailing newline character
                if self.coder:
                    line = self.coder.decode(line)
                yield line



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

import atexit
import codecs
import typing as t

from cython.operator cimport preincrement as preinc
from libcpp.string cimport string

from resiliparse_inc.lexbor cimport lxb_char_t, lxb_status_t, lxb_html_encoding_t, lxb_html_encoding_entry_t, \
    lxb_html_encoding_determine, lxb_html_encoding_meta_entry, lxb_html_encoding_init, lxb_html_encoding_destroy, \
    LXB_STATUS_OK
from resiliparse_inc.uchardet cimport uchardet_new, uchardet_delete, uchardet_handle_data, \
    uchardet_data_end, uchardet_reset, uchardet_get_charset


# Encoding name and label map according to https://encoding.spec.whatwg.org/#names-and-labels
# Differences:
#   * ISO-8859-8-I name replaced with ISO-8859-8
#   * WINDOWS-874 name replaced with ISO-8859-11
#   * x-mac-cyrillic is unsupported
#   * x-user-defined is unsupported
#   * No "replacement" mapping for 7-bit versions of ISO/IEC 2022
cdef dict __enc_html5_map = {'866': 'cp866', 'ansi_x3.4-1968': 'cp1252', 'arabic': 'iso8859-6', 'ascii': 'cp1252',
                             'asmo-708': 'iso8859-6', 'big5': 'big5', 'big5-hkscs': 'big5', 'chinese': 'gbk',
                             'cn-big5': 'big5', 'cp1250': 'cp1250', 'cp1251': 'cp1251', 'cp1252': 'cp1252',
                             'cp1253': 'cp1253', 'cp1254': 'cp1254', 'cp1255': 'cp1255', 'cp1256': 'cp1256',
                             'cp1257': 'cp1257', 'cp1258': 'cp1258', 'cp819': 'cp1252', 'cp866': 'cp866',
                             'csbig5': 'big5', 'cseuckr': 'euc_kr', 'cseucpkdfmtjapanese': 'euc_jp', 'csgb2312': 'gbk',
                             'csibm866': 'cp866', 'csiso2022jp': 'iso2022_jp', 'csiso58gb231280': 'gbk',
                             'csiso88596e': 'iso8859-6', 'csiso88596i': 'iso8859-6', 'csiso88598e': 'iso8859-8',
                             'csiso88598i': 'iso8859-8', 'csisolatin1': 'cp1252', 'csisolatin2': 'iso8859-2',
                             'csisolatin3': 'iso8859-3', 'csisolatin4': 'iso8859-4', 'csisolatin5': 'cp1254',
                             'csisolatin6': 'iso8859-10', 'csisolatin9': 'iso8859-15', 'csisolatinarabic': 'iso8859-6',
                             'csisolatincyrillic': 'iso8859-5', 'csisolatingreek': 'iso8859-7',
                             'csisolatinhebrew': 'iso8859-8', 'cskoi8r': 'koi8-r', 'csksc56011987': 'euc_kr',
                             'csmacintosh': 'mac-roman', 'csshiftjis': 'shift_jis', 'csunicode': 'utf-16-le',
                             'cyrillic': 'iso8859-5', 'dos-874': 'iso8859-11', 'ecma-114': 'iso8859-6',
                             'ecma-118': 'iso8859-7', 'elot_928': 'iso8859-7', 'euc-jp': 'euc_jp', 'euc-kr': 'euc_kr',
                             'euc_jp': 'euc_jp', 'euc_kr': 'euc_kr', 'gb18030': 'gb18030', 'gb2312': 'gbk',
                             'gb_2312': 'gbk', 'gb_2312-80': 'gbk', 'gbk': 'gbk', 'greek': 'iso8859-7',
                             'greek8': 'iso8859-7', 'hebrew': 'iso8859-8', 'ibm819': 'cp1252', 'ibm866': 'cp866',
                             'iso-10646-ucs-2': 'utf-16-le', 'iso-2022-jp': 'iso2022_jp', 'iso-8859-1': 'cp1252',
                             'iso-8859-10': 'iso8859-10', 'iso-8859-11': 'iso8859-11', 'iso-8859-13': 'iso8859-13',
                             'iso-8859-14': 'iso8859-14', 'iso-8859-15': 'iso8859-15', 'iso-8859-16': 'iso8859-16',
                             'iso-8859-2': 'iso8859-2', 'iso-8859-3': 'iso8859-3', 'iso-8859-4': 'iso8859-4',
                             'iso-8859-5': 'iso8859-5', 'iso-8859-6': 'iso8859-6', 'iso-8859-6-e': 'iso8859-6',
                             'iso-8859-6-i': 'iso8859-6', 'iso-8859-7': 'iso8859-7', 'iso-8859-8': 'ISO-8859-8',
                             'iso-8859-8-e': 'iso8859-8', 'iso-8859-8-i': 'iso8859-8', 'iso-8859-9': 'cp1254',
                             'iso-ir-100': 'cp1252', 'iso-ir-101': 'iso8859-2', 'iso-ir-109': 'iso8859-3',
                             'iso-ir-110': 'iso8859-4', 'iso-ir-126': 'iso8859-7', 'iso-ir-127': 'iso8859-6',
                             'iso-ir-138': 'iso8859-8', 'iso-ir-144': 'iso8859-5', 'iso-ir-148': 'cp1254',
                             'iso-ir-149': 'euc_kr', 'iso-ir-157': 'iso8859-10', 'iso-ir-58': 'gbk',
                             'iso2022_jp': 'iso2022_jp', 'iso8859-1': 'cp1252', 'iso8859-10': 'iso8859-10',
                             'iso8859-11': 'iso8859-11', 'iso8859-13': 'iso8859-13', 'iso8859-14': 'iso8859-14',
                             'iso8859-15': 'iso8859-15', 'iso8859-16': 'iso8859-16', 'iso8859-2': 'iso8859-2',
                             'iso8859-3': 'iso8859-3', 'iso8859-4': 'iso8859-4', 'iso8859-5': 'iso8859-5',
                             'iso8859-6': 'iso8859-6', 'iso8859-7': 'iso8859-7', 'iso8859-8': 'iso8859-8',
                             'iso8859-9': 'cp1254', 'iso88591': 'cp1252', 'iso885910': 'iso8859-10',
                             'iso885911': 'iso8859-11', 'iso885913': 'iso8859-13', 'iso885914': 'iso8859-14',
                             'iso885915': 'iso8859-15', 'iso88592': 'iso8859-2', 'iso88593': 'iso8859-3',
                             'iso88594': 'iso8859-4', 'iso88595': 'iso8859-5', 'iso88596': 'iso8859-6',
                             'iso88597': 'iso8859-7', 'iso88598': 'iso8859-8', 'iso88599': 'cp1254',
                             'iso_8859-1': 'cp1252', 'iso_8859-15': 'iso8859-15', 'iso_8859-1:1987': 'cp1252',
                             'iso_8859-2': 'iso8859-2', 'iso_8859-2:1987': 'iso8859-2', 'iso_8859-3': 'iso8859-3',
                             'iso_8859-3:1988': 'iso8859-3', 'iso_8859-4': 'iso8859-4', 'iso_8859-4:1988': 'iso8859-4',
                             'iso_8859-5': 'iso8859-5', 'iso_8859-5:1988': 'iso8859-5', 'iso_8859-6': 'iso8859-6',
                             'iso_8859-6:1987': 'iso8859-6', 'iso_8859-7': 'iso8859-7', 'iso_8859-7:1987': 'iso8859-7',
                             'iso_8859-8': 'iso8859-8', 'iso_8859-8:1988': 'iso8859-8', 'iso_8859-9': 'cp1254',
                             'iso_8859-9:1989': 'cp1254', 'koi': 'koi8-r', 'koi8': 'koi8-r', 'koi8-r': 'koi8-r',
                             'koi8-ru': 'koi8-u', 'koi8-u': 'koi8-u', 'koi8_r': 'koi8-r', 'korean': 'euc_kr',
                             'ks_c_5601-1987': 'euc_kr', 'ks_c_5601-1989': 'euc_kr', 'ksc5601': 'euc_kr',
                             'ksc_5601': 'euc_kr', 'l1': 'cp1252', 'l2': 'iso8859-2', 'l3': 'iso8859-3',
                             'l4': 'iso8859-4', 'l5': 'cp1254', 'l6': 'iso8859-10', 'l9': 'iso8859-15',
                             'latin1': 'cp1252', 'latin2': 'iso8859-2', 'latin3': 'iso8859-3', 'latin4': 'iso8859-4',
                             'latin5': 'cp1254', 'latin6': 'iso8859-10', 'logical': 'iso8859-8', 'mac': 'mac-roman',
                             'mac-roman': 'mac-roman', 'macintosh': 'mac-roman', 'ms932': 'shift_jis',
                             'ms_kanji': 'shift_jis', 'shift-jis': 'shift_jis', 'shift_jis': 'shift_jis',
                             'sjis': 'shift_jis', 'sun_eu_greek': 'iso8859-7', 'tis-620': 'iso8859-11',
                             'ucs-2': 'utf-16-le', 'unicode': 'utf-16-le', 'unicode-1-1-utf-8': 'utf-8',
                             'unicode11utf8': 'utf-8', 'unicode20utf8': 'utf-8', 'unicodefeff': 'utf-16-le',
                             'unicodefffe': 'utf-16-be', 'us-ascii': 'cp1252', 'utf-16': 'utf-16-le',
                             'utf-16-be': 'utf-16-be', 'utf-16-le': 'utf-16-le', 'utf-16be': 'utf-16-be',
                             'utf-16le': 'utf-16-le', 'utf-8': 'utf-8', 'utf8': 'utf-8', 'visual': 'iso8859-8',
                             'windows-1250': 'cp1250', 'windows-1251': 'cp1251', 'windows-1252': 'cp1252',
                             'windows-1253': 'cp1253', 'windows-1254': 'cp1254', 'windows-1255': 'cp1255',
                             'windows-1256': 'cp1256', 'windows-1257': 'cp1257', 'windows-1258': 'cp1258',
                             'windows-31j': 'shift_jis', 'windows-874': 'iso8859-11', 'windows-949': 'euc_kr',
                             'x-cp1250': 'cp1250', 'x-cp1251': 'cp1251', 'x-cp1252': 'cp1252', 'x-cp1253': 'cp1253',
                             'x-cp1254': 'cp1254', 'x-cp1255': 'cp1255', 'x-cp1256': 'cp1256', 'x-cp1257': 'cp1257',
                             'x-cp1258': 'cp1258', 'x-euc-jp': 'euc_jp', 'x-gbk': 'gbk', 'x-mac-roman': 'mac-roman',
                             'x-sjis': 'shift_jis', 'x-unicode20utf8': 'utf-8', 'x-x-big5': 'big5'}


cdef class EncodingDetector:
    """
    Universal character encoding detector based on `uchardet`.

    `uchardet` is a C wrapper and a continuation of Mozilla's `Universal Charset Detector` library.
    """

    def __cinit__(self):
        self.d = uchardet_new()

    cpdef void update(self, const string& data):
        """
        update(self, data)

        Update charset detector with more data.

        The detector will shortcut processing when it has enough data to reach certainty, so you don't
        need to worry too much about limiting the input data.

        :param data: input data
        :type data: bytes
        """
        uchardet_handle_data(self.d, data.data(), data.size())

    cpdef str encoding(self, bint html5_compatible=True):
        """
        encoding(self, html5_compatible=True)

        Get a Python-compatible name of the encoding that was detected and reset the detector.

        By default, the detected encoding is remapped based on the `WHATWG encoding specification
        <https://encoding.spec.whatwg.org/#names-and-labels>`_, which is primarily suitable for web
        content. To disable this behaviour, set ``html5_compatible=False``. For more information,
        see: :func:`map_encoding_to_html5`.

        If WHATWG remapping is enabled, UTF-8 is returned as a fallback encoding. Otherwise, the
        method returns ``None`` on failure to detect the encoding.

        :param html5_compatible: Remap encoding names according to WHATWG
        :type html5_compatible: bool
        :return: detected encoding or ``None`` on failure
        :rtype: str or None
        """
        uchardet_data_end(self.d)
        cdef str enc = uchardet_get_charset(self.d).decode()
        uchardet_reset(self.d)

        if html5_compatible:
            enc = map_encoding_to_html5(enc)
        else:
            try:
                enc = codecs.lookup(enc).name
            except LookupError:
                return None

        return enc

    def __dealloc__(self):
        if self.d != NULL:
            uchardet_delete(self.d)
            self.d = NULL


cdef EncodingDetector __chardet = None


@atexit.register
def __chardet_exit():
    global __chardet
    __chardet = None


cpdef str detect_encoding(bytes data, size_t max_len=4096, bint html5_compatible=True, bint from_html_meta=False):
    """
    detect_encoding(data, max_len=4096, html5_compatible=True, from_html_meta=False)

    Detect the encoding of a byte string. This is a convenience wrapper around :class:`EncodingDetector`
    that uses a single global instance.

    The string that is passed to the :class:`EncodingDetector` will be no longer than ``max_len``
    bytes to prevent slow-downs and keep memory usage low. If the string is longer than this limit, only
    the ``max_len / 2`` bytes from the start and from the end of the string will be used. This is a tradeoff
    between performance and accuracy. If you need higher accuracy, increase the limit to feed more data
    into the :class:`EncodingDetector`.

    The :class:`EncodingDetector` relies on `uchardet` as its encoding detection engine. If the
    input string is an HTML document, you can also use the available information from the HTML meta charset
    tag instead. With ``from_html_meta=True``, :func:`detect_encoding` will try to use the charset meta
    tag in the HTML string if one is available and ASCII-readable within the first 1024 bytes. If this fails,
    it will fall back to auto-detection with `uchardet`.

    By default, the detected encoding name is remapped according to the `WHATWG encoding specification
    <https://encoding.spec.whatwg.org/#names-and-labels>`_, which is primarily suitable for web content.
    To disable this behaviour, set ``html5_compatible=False``. For more information, see:
    :func:`map_encoding_to_html5`. Encodings detected from HTML meta tags are always remapped, no matter the
    value of ``html5_compatible``, to ensure valid encoding names.

    If WHATWG remapping is enabled, UTF-8 is returned as a fallback encoding. Otherwise, the method returns
    ``None`` on failure to detect the encoding.

    :param data: input string for which to detect the encoding
    :type data: bytes
    :param max_len: maximum number of bytes to feed to detector (0 for no limit)
    :type max_len: int
    :param html5_compatible: Remap encoding names according to WHATWG
    :type html5_compatible: bool
    :param from_html_meta: if string is an HTML document, use meta tag info
    :type from_html_meta: bool
    :return: detected encoding
    :rtype: str
    """
    cdef lxb_status_t status
    cdef lxb_html_encoding_t html_enc
    cdef lxb_html_encoding_entry_t* html_enc_entry
    cdef size_t meta_peek_size = min(len(data), 1024)
    if from_html_meta:
        if lxb_html_encoding_init(&html_enc) == LXB_STATUS_OK and lxb_html_encoding_determine(
                &html_enc, <lxb_char_t*>data, <lxb_char_t*>data + meta_peek_size) == LXB_STATUS_OK:
            html_enc_entry = lxb_html_encoding_meta_entry(&html_enc, 0)
            html_enc_name = ''
            if html_enc_entry != NULL:
                html_enc_name = html_enc_entry.name[:html_enc_entry.end - html_enc_entry.name].decode(errors='ignore')
            lxb_html_encoding_destroy(&html_enc, False)
            html_enc_name = map_encoding_to_html5(html_enc_name, fallback_utf8=False)
            if html_enc_name:
                return html_enc_name

    if max_len != 0 and <size_t>len(data) > max_len:
        data = data[:(max_len + 1) // 2] + data[-((<ssize_t>max_len + 1) // 2):]

    global __chardet
    if __chardet is None:
        __chardet = EncodingDetector.__new__(EncodingDetector)
    __chardet.update(<string> data)
    return __chardet.encoding(html5_compatible)


cpdef str map_encoding_to_html5(str encoding, bint fallback_utf8=True):
    """
    map_encoding_to_html5(encoding, fallback_utf8=True)

    Map an encoding name to a subset of names allowed by the HTML5 standard.

    This function will remap the given name according to the mapping definition given in
    `Section 4.2 <https://encoding.spec.whatwg.org/#names-and-labels>`_ of the WHATWG encoding
    specification. The returned value will always be a valid Python encoding name, but the
    supplied input name does not necessarily have to be.

    The WHATWG mapping is designed to boil down the many possible encoding names to a smaller
    subset of canonical names while taking into account common encoding mislabelling practices.
    The main purpose of this function is to map encoding names extracted from HTTP headers or
    websites to their canonical names, but it also makes sense to apply the mapping to an
    auto-detected encoding name, since it remaps some encodings based on observed practices on
    the web, such as the mapping from ISO-8859-1 to Windows-1252, which is more likely to be
    correct, even if both options are possible. :meth:`EncodingDetector.encoding` already remaps
    its detected encodings to the WHATWG set by default.

    The mapping does not involve Python's encoding alias names, but instead uses an adjusted
    WHATWG mapping. Inputs not defined in this mapping are remapped to UTF-8. Hence, the function
    always produces a valid output, but the mapped encoding is not guaranteed to be compatible
    with the original encoding. Use :func:`bytes_to_str` to avoid decoding errors. You can also
    set ``fallback_utf8=False`` to return ``None`` instead if the supplied encoding is unknown.

    The adjusted encoding mapping differs from the WHATWG spec in the following details:

      * ISO-8859-8-I name replaced with ISO-8859-8
      * WINDOWS-874 name replaced with ISO-8859-11
      * x-mac-cyrillic is unsupported
      * x-user-defined is unsupported
      * No "replacement" mapping for 7-bit versions of ISO/IEC 2022

    :param encoding: input encoding name
    :type encoding: str
    :param fallback_utf8: Whether to fall back to UTF-8 or return ``None`` for unknown encodings
    :type fallback_utf8: bool
    :return: mapped output encoding name
    :rtype: str or None
    """
    return __enc_html5_map.get(encoding.strip().casefold(), 'utf-8' if fallback_utf8 else None)


cdef inline str __map_utf(str enc, bytes data, bint strip):
    if not strip:
        return enc

    if enc == 'utf-8' and data.startswith(codecs.BOM_UTF8):
        return 'utf-8-sig'
    if enc.startswith('utf-16-') and (data.startswith(codecs.BOM_UTF16_LE) or data.startswith(codecs.BOM_UTF16_BE)):
        return 'utf-16'
    if enc.startswith('utf-32-') and (data.startswith(codecs.BOM_UTF32_LE) or data.startswith(codecs.BOM_UTF32_BE)):
        return 'utf-32'

    return enc


# noinspection PyTypeChecker
cpdef str bytes_to_str(bytes data, str encoding='utf-8', str errors='ignore',
                       fallback_encodings=('utf-8', 'cp1252'), bint strip_bom=True):
    """
    bytes_to_str(data, encoding='utf-8', errors='ignore', fallback_encodings=('utf-8', 'cp1252'), strip_bom=True)

    Helper for decoding a byte string into a unicode string using a given encoding.
    This encoding should be determined beforehand, e.g., with :func:`detect_encoding`.

    :func:`bytes_to_str` tries to decode the byte string with ``encoding``. If that
    fails, it will fall back to UTF-8 and Windows-1252 (or whichever encodings where
    given in ``fallback_encodings``). If all fallbacks fail as well, the string will be
    double-decoded with ``encoding`` and invalid characters will be treated according
    to ``errors``, which has the same options as for :meth:`bytes.decode` (i.e.,
    ``"ignore"`` or ``"replace"``). The double-decoding step ensures that the resulting
    string is sane and can be re-encoded without errors.

    This function also takes care to strip BOMs from the beginning of the string if
    ``strip_bom=True``.

    :param data: input byte string
    :type data: bytes
    :param encoding: desired encoding
    :type encoding: str
    :param errors: error handling for invalid characters
    :type errors: str
    :param fallback_encodings: list of fallback encodings to try if the primary encoding fails
    :type fallback_encodings: t.Iterable[str]
    :param strip_bom: strip BOM sequences from beginning of the string
    :type strip_bom: bool
    :return: decoded string
    :rtype: str
    """

    encoding = codecs.lookup(encoding).name

    for i, e in enumerate((encoding, *fallback_encodings)):
        if i > 0:
            e = codecs.lookup(e).name
            if e == encoding:
                # No need to try that again
                continue

        try:
            return data.decode(__map_utf(e, data, strip_bom))
        except UnicodeDecodeError:
            pass

    return data.decode(__map_utf(encoding, data, strip_bom), errors=errors).encode(errors=errors).decode()


cpdef str detect_mime(bytes data, float max_unprintable=0.05):
    """
    detect_mime(data, max_unprintable=0.05)
    
    Try to detect common internet MIME types based on the initial magic byte sequence of ``data``.
    
    The check is very basic and only checks the starting bytes as well as the number of unprintable
    bytes. It does not replace a full-blown MIME type detection engine like Apache Tika at the moment.
    
    :param data: input bytes
    :type data: bytes
    :param max_unprintable: maximum allowable ratio of unprintable characters for text
    :type max_unprintable: float
    :return: detected MIME type
    :rtype: str
    """

    cdef size_t i
    cdef bytes data_strip = data[:128].lstrip()
    for i in range(MIME_BYTES.size()):
        if data.startswith(MIME_BYTES[i].magic_bytes) or data_strip.startswith(MIME_BYTES[i].magic_bytes):
            if MIME_BYTES[i].mime_type in [b'text/html', b'application/xml']:
                if data_strip.find(b'"-//W3C//DTD XHTML') > -1:
                    return 'application/xhtml+xml'
                if MIME_BYTES[i].mime_type == b'application/xml':
                    if data_strip.find(b'<svg ') > -1 or data_strip.find(b'"-//W3C//DTD SVG') > -1:
                        return 'image/svg+xml'
            return MIME_BYTES[i].mime_type.decode()

    cdef size_t unprintable = 0
    for i in range(<size_t>len(data)):
        if 0x00 <= data[i] <= 0x1f or 0x7f <= data[i] <= 0xff:
            preinc(unprintable)
        if unprintable > 0 and unprintable > <size_t>(len(data) * max_unprintable):
            return 'application/octet-stream'

    return 'text/plain'



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

import apache_beam as beam
from apache_beam.utils.windowed_value import WindowedValue
import apache_beam.typehints as t
from elasticsearch import exceptions as es_exc, Elasticsearch
from elasticsearch.helpers import BulkIndexError, streaming_bulk


logger = logging.getLogger()


class ElasticsearchBulkIndex(beam.PTransform):
    """
    Bulk-index documents to Elasticsearch.

    Expects bulk index actions that can be passed to :func:`elasticsearch.helpers.streaming_bulk`.
    Helpers for creating such actions are given with :func:`index_action`, :func:`update_action`,
    and :func:`delete_action`.

    If the records are KV pairs, the key will be used as the document ``_id`` and the value is
    expected to be the document that is to be indexed as a dict. In this case, fields starting with
    underscores will be discarded and ``default_index`` must be set.

    Returns the document IDs of successfully indexed documents.

    :param es_args: Elasticsearch client arguments
    :param default_index: default index name if none is given in the index action
    :param update: do a bulk UPDATE instead of a bulk INDEX (requires documents to exist)
    :param parallelism: reshuffle to achieve the desired level of parallelism
    :param buffer_size: internal buffer size
    :param chunk_size: indexing chunk size
    :param max_retries: maximum number of retries on recoverable failures
    :param initial_backoff: initial retry backoff
    :param max_backoff: maximum retry backoff
    :param request_timeout: Elasticsearch request timeout
    :param ignore_400: ignore HTTP 400 errors (skip unindexable documents)
    :param dry_run: discard documents and do not actually index them
    """

    def __init__(self,
                 es_args: t.Dict[str, t.Any],
                 default_index: t.Optional[str] = None,
                 update: bool = False,
                 parallelism: int = None,
                 buffer_size: int = 3200,
                 chunk_size: int = 800,
                 max_retries: int = 10,
                 initial_backoff: float = 2,
                 max_backoff: float = 600,
                 request_timeout: int = 240,
                 ignore_400: bool = True,
                 dry_run: bool = False):

        super().__init__()
        self._pardo = _ElasticsearchBulkIndex(es_args=es_args,
                                              default_index=default_index,
                                              update=update,
                                              buffer_size=buffer_size,
                                              chunk_size=chunk_size,
                                              max_retries=max_retries,
                                              initial_backoff=initial_backoff,
                                              max_backoff=max_backoff,
                                              request_timeout=request_timeout,
                                              ignore_400=ignore_400,
                                              dry_run=dry_run)
        self._parallelism = parallelism

    def expand(self, pcoll):
        if self._parallelism:
            pcoll |= beam.Reshuffle(self._parallelism)
        return pcoll | beam.ParDo(self._pardo)


# noinspection PyAbstractClass
class _ElasticsearchBulkIndex(beam.DoFn):
    def __init__(self,
                 es_args,
                 default_index,
                 update,
                 buffer_size,
                 chunk_size,
                 max_retries,
                 initial_backoff,
                 max_backoff,
                 request_timeout,
                 ignore_400,
                 dry_run):
        super().__init__()

        self.buffer_size = buffer_size
        self.buffer = []

        self.es_args = es_args
        self.default_index = default_index
        self.update = update
        self.client = None
        self.bulk_args = dict(
            chunk_size=chunk_size,
            # We handle errors ourselves, so don't raise and don't let Elasticsearch retry
            # HTTP 429 errors, since that would mess with the result order.
            raise_on_exception=False,
            raise_on_error=False,
            max_retries=0,
            request_timeout=request_timeout,
            yield_ok=True
        )
        self.max_retries = max_retries
        self.initial_backoff = initial_backoff
        self.max_backoff = max_backoff
        self.ignore_400 = ignore_400
        self.dry_run = dry_run

    def setup(self):
        self.client = Elasticsearch(**self.es_args)

    def teardown(self):
        assert len(self.buffer) == 0
        if self.client:
            self.client.transport.close()

    # noinspection PyIncorrectDocstring
    def process(self,
                element: t.Union[t.KV[str, t.Dict[str, t.Any]], t.Dict[str, t.Any]],
                timestamp=beam.DoFn.TimestampParam,
                window=beam.DoFn.WindowParam):
        """
        Add element to index buffer.

        Yields the document IDs of successfully indexed document on buffer flush.

        :param element: input index action or KV pair containing the document ID and the document values.
        :return: iterable of indexed document IDs
        """

        if type(element) is tuple and len(element) == 2:
            if not self.default_index:
                raise RuntimeError('Got a KV pair, but default index is not set.')
            doc_id, element = element
            element = index_action(doc_id, self.default_index, element)

        if self.dry_run:
            yield element.get('_id')
            return

        self.buffer.append((element, timestamp, window))
        if len(self.buffer) >= self.buffer_size:
            yield from self._flush_buffer()

    def finish_bundle(self):
        if len(self.buffer) > 0:
            yield from self._flush_buffer(window_values=True)

    def _flush_buffer(self, window_values=False):
        if not self.buffer:
            return

        retry_count = 0
        errors = []
        self.buffer.sort(key=lambda x: x[0].get('_id', ''))

        try:
            while retry_count < self.max_retries + 1:
                try:
                    errors = []
                    to_retry = []

                    buf_gen = (e[0] for e in self.buffer)
                    for i, (ok, info) in enumerate(streaming_bulk(self.client, buf_gen, **self.bulk_args)):
                        info_val = list(info.values())[0]
                        if not ok:
                            errors.append(info)
                            if info_val.get('status') != 400 or not self.ignore_400:
                                # 400 errors are persistent, no point in retrying them
                                to_retry.append(self.buffer[i])
                            logger.error('Failed to index document (attempt %i/%i): %s',
                                         retry_count + 1, self.max_retries + 1, info)
                        else:
                            if window_values:
                                yield WindowedValue(info_val['_id'], self.buffer[i][1], [self.buffer[i][2]])
                            else:
                                yield info_val['_id']

                    if not to_retry:
                        return

                    self.buffer = to_retry

                except es_exc.TransportError as e:
                    logger.error('Elasticsearch transport error (attempt %i/%i): %s',
                                 retry_count + 1, self.max_retries + 1, e.message)
                    if retry_count >= self.max_retries:
                        raise e

                if retry_count >= self.max_retries:
                    break
                logger.error('Retrying with exponential backoff in %i seconds...',
                             self.initial_backoff * (2 ** retry_count))
                time.sleep(min(self.max_backoff, self.initial_backoff * (2 ** retry_count)))
                retry_count += 1
        finally:
            self.buffer.clear()

        raise BulkIndexError("%i document(s) failed to index." % len(errors), errors)


def index_action(doc_id: str, index: t.Optional[str], doc: t.Dict[str, t.Any]):
    """
    Build an index bulk action.

    :param doc_id: document ID (``None`` for using autogenerated IDs)
    :param index: index name
    :param doc: document as dict
    :return: index action dict
    """
    d = {
        '_op_type': 'index',
        '_index': index,
        '_id': doc_id,
        **{k: v for k, v in doc.items() if not k.startswith('_')}
    }
    if d['_id'] is None:
        del d['_id']
    return d


def update_action(doc_id: str, index: str, partial_doc: t.Dict[str, t.Any]):
    """
    Build an update bulk action.

    :param doc_id: document ID
    :param index: index name
    :param partial_doc: partial update document as dict
    :return: index action dict
    """
    return {
        '_op_type': 'update',
        '_index': index,
        '_id': doc_id,
        'doc': {k: v for k, v in partial_doc.items() if not k.startswith('_')}
    }


def delete_action(doc_id: str, index: str):
    """
    Build a delete bulk action.

    :param doc_id: document ID
    :param index: index name
    :return: index action dict
    """
    return {
        '_op_type': 'delete',
        '_index': index,
        '_id': doc_id
    }


def ensure_index(client: Elasticsearch,
                 index_name: str,
                 index_settings: t.Dict[str, str] = None,
                 mapping: t.Dict[str, str] = None):     # pragma: no cover
    """
    Helper function to ensure an index exists.

    If the index does not exist, it will be created with the given mapping and settings.

    :param client: Elasticsearch client instance
    :param index_name: index name
    :param index_settings: index settings dict
    :param mapping: index mapping dict
    """
    index_settings = index_settings or {}
    mapping = mapping or {}

    if not client.indices.exists(index=index_name):
        client.indices.create(index=index_name, body=dict(
            settings=index_settings,
            mappings=mapping
        ))
